import json

import httpx
from httpx_sse import aconnect_sse, connect_sse


class _async_http_client:
    def __init__(self, api_key: str, api_url, timeout):
        self.session = httpx.AsyncClient(
            base_url=api_url,
            headers={"X-OPPER-API-KEY": f"{api_key}"},
            timeout=httpx.Timeout(timeout),
        )

    async def do_request(self, method: str, path: str, **kwargs):
        return await self.session.request(method, path, **kwargs)

    async def stream(self, method: str, path: str, **kwargs):
        async with aconnect_sse(self.session, method, path, **kwargs) as event_source:
            async for sse in event_source.aiter_sse():
                yield json.loads(sse.data)


class _http_client:
    def __init__(self, api_key: str, api_url: str, timeout):
        self.session = httpx.Client(
            base_url=api_url,
            headers={"X-OPPER-API-KEY": f"{api_key}"},
            timeout=httpx.Timeout(timeout),
        )

    def do_request(self, method: str, path: str, **kwargs):
        return self.session.request(method, path, **kwargs)

    def stream(self, method: str, path: str, **kwargs):
        with connect_sse(
            self.session,
            method,
            path,
            **kwargs,
        ) as event_source:
            for sse in event_source.iter_sse():
                yield json.loads(sse.data)



from typing import List, Optional

from opperai._http_clients import _async_http_client
from opperai.types.exceptions import APIError
from opperai.types.indexes import (
    Document,
    Filter,
    Index,
    RetrievalResponse,
)
from ._indexes import RetrieveQuery


class AsyncIndexes:
    def __init__(self, http_client: _async_http_client):
        self.http_client = http_client

    async def create(self, name: str) -> Index:
        response = await self.http_client.do_request(
            "POST",
            "/v1/indexes",
            json={"name": name},
        )
        if response.status_code != 200:
            raise APIError(f"Failed to create index with status {response.status_code}")
        return Index.model_validate(response.json())

    async def delete(self, id: int) -> bool:
        response = await self.http_client.do_request(
            "DELETE",
            f"/v1/indexes/{id}",
        )
        if response.status_code == 404:
            return False
        if response.status_code != 200:
            raise APIError(f"Failed to delete index with status {response.status_code}")
        return True

    async def get(self, id: int = None, name: str = None) -> Index:
        if id is None and name is None:
            raise ValueError("Either id or name must be provided")
        if id is not None and name is not None:
            raise ValueError("Only one of id or name should be provided")

        if id is not None:
            return await self._get_by_id(id)
        if name is not None:
            return await self._get_by_name(name)

    async def _get_by_id(self, id: int) -> Index:
        response = await self.http_client.do_request(
            "GET",
            f"/v1/indexes/{id}",
        )
        if response.status_code == 404:
            return None
        if response.status_code != 200:
            raise APIError(
                f"Failed to get index {id} with status {response.status_code}"
            )
        return Index.model_validate(response.json())

    async def _get_by_name(self, name: str) -> Index:
        indexes = await self.list()
        for index in indexes:
            if index.name == name:
                return index
        return None

    async def list(self) -> List[Index]:
        response = await self.http_client.do_request(
            "GET",
            "/v1/indexes",
        )
        if response.status_code != 200:
            raise APIError(f"Failed to list indexes with status {response.status_code}")
        return [Index.model_validate(item) for item in response.json()]

    async def upload_file(self, id: int, file_path: str, **kwargs):
        # Get upload URL
        upload_url_response = await self.http_client.do_request(
            "GET",
            f"/v1/indexes/{id}/upload_url",
            params={"filename": file_path.split("/")[-1]},
        )
        if upload_url_response.status_code != 200:
            raise APIError(
                f"Failed to get upload URL with status {upload_url_response.status_code}"
            )
        upload_url_data = upload_url_response.json()

        # Upload file
        with open(file_path, "rb") as file:
            files = {"file": (file_path.split("/")[-1], file)}
            upload_response = await self.http_client.do_request(
                "POST",
                upload_url_data["url"],
                files=files,
                data=upload_url_data["fields"],
            )
            if upload_response.status_code not in [200, 204]:
                raise APIError(
                    f"Failed to upload file with status {upload_response.status_code}"
                )

        # Register file
        register_file_response = await self.http_client.do_request(
            "POST",
            f"/v1/indexes/{id}/register_file",
            json={"uuid": upload_url_data["uuid"], **kwargs},
        )
        if register_file_response.status_code != 200:
            raise APIError(
                f"Failed to register file with status {register_file_response.status_code}"
            )

        return register_file_response.json()

    async def index(self, id: int, doc: Document) -> Document:
        response = await self.http_client.do_request(
            "POST",
            f"/v1/indexes/{id}/index",
            json=doc.model_dump(),
        )
        if response.status_code != 200:
            raise APIError(f"Failed to add document with status {response.status_code}")
        return Document.model_validate(response.json())

    async def retrieve(
        self, id: int, query: str, k: int, filters: Optional[List[Filter]] = None
    ) -> List[RetrievalResponse]:
        response = await self.http_client.do_request(
            "POST",
            f"/v1/indexes/{id}/query",
            json=RetrieveQuery(q=query, k=k, filters=filters).model_dump(),
        )
        if response.status_code != 200:
            raise APIError(
                f"Failed to retrieve index {id} with status {response.status_code}"
            )
        return [RetrievalResponse.model_validate(item) for item in response.json()]



import inspect
from typing import Dict, List, Union, get_args, get_origin, get_type_hints

from pydantic import BaseModel


def type_to_json_schema(type_hint):
    schema = _type_to_json_schema(type_hint)
    schema, defs = _lift_defs(schema, {})
    schema["$defs"] = defs
    return schema


def _lift_defs(schema: Dict, defs: Dict = {}):
    if "$defs" in schema:
        for k, v in schema["$defs"].items():
            defs[k] = v
        schema.pop("$defs")

    for k, v in schema.items():
        if isinstance(v, dict):
            _lift_defs(v, defs)
        elif isinstance(v, list):
            for item in v:
                if isinstance(item, dict):
                    _lift_defs(item, defs)

    return schema, defs


def _type_to_json_schema(type_hint):
    """Convert a Python type to a JSON schema."""
    if type_hint == int:
        return {"type": "integer"}
    elif type_hint == str:
        return {"type": "string"}
    elif type_hint == float:
        return {"type": "number"}
    elif type_hint == bool:
        return {"type": "boolean"}
    elif inspect.isclass(type_hint) and issubclass(type_hint, BaseModel):
        return type_hint.model_json_schema()
    elif hasattr(type_hint, "__origin__"):
        # Handling generic types (List, Dict, etc.)
        if get_origin(type_hint) == Union and type(None) in get_args(type_hint):
            return _type_to_json_schema(get_args(type_hint)[0])  # for Optional
        elif type_hint.__origin__ in (List, list):
            item_type = get_args(type_hint)[0]
            if inspect.isclass(item_type):
                return {"type": "array", "items": _type_to_json_schema(item_type)}
            if get_origin(item_type) == List:
                return {"type": "array", "items": _type_to_json_schema(item_type)}
            elif inspect.isclass(item_type) and issubclass(item_type, BaseModel):
                return {"type": "array", "items": item_type.model_json_schema()}
            else:
                return {"type": "array", "items": _type_to_json_schema(item_type)}
        elif get_origin(type_hint) == Dict:
            return {"type": "object"}
        else:
            raise NotImplementedError(
                f"Type {type_hint} is not supported for JSON schema generation.{get_origin(type_hint)}"
            )
    else:
        return {"type": "object"}


def get_input_schema(func):
    """Generate input JSON schema from a function signature."""
    sig = inspect.signature(func)
    type_hints = get_type_hints(func)

    schema = {"type": "object", "properties": {}, "required": []}

    for param in sig.parameters.values():
        param_type = type_hints.get(
            param.name, str
        )  # Default to string if type is not specified
        schema["properties"][param.name] = type_to_json_schema(param_type)
        if param.default is inspect.Parameter.empty:
            schema["required"].append(param.name)

    if not schema["properties"]:
        return {
            "type": "object",
            "description": "No input parameters",
            "properties": {},
            "additionalProperties": False,
        }

    return schema


def get_output_schema(func):
    """Generate output JSON schema from a function signature."""
    return_type = get_type_hints(func).get("return", None)
    if return_type:
        schema = type_to_json_schema(return_type)
    else:
        schema = {"type": "unknown"}

    return schema



import os
import uuid
from unittest.mock import MagicMock, patch
from datetime import datetime

from uuid import uuid4
from opperai.types.spans import Span
from opperai import Client

from opperai import trace


@patch("opperai._http_clients._http_client.do_request")
def test_decorator(mock_do_request):
    mock_do_request.side_effect = [
        MagicMock(status_code=200, json=lambda: {"uuid": str(uuid.uuid4())}),
        MagicMock(status_code=200, json=lambda: {"uuid": str(uuid.uuid4())}),
    ]

    @trace(client=Client(api_key="temporary", api_url="temporary"))
    def something(text: str, target_language: str) -> str:
        return "Hola"

    assert something("Hello", "es") == "Hola"



import pytest
from pydantic import ValidationError
from opperai.types import validate_id_xor_path, SpanFeedback


def test_id_xor_path():
    @validate_id_xor_path
    def test_function(id=None, path=None):
        return id, path

    assert test_function(id=1) == (1, None)

    assert test_function(path="test") == (None, "test")

    with pytest.raises(ValueError):
        test_function(id=1, path="test")

    with pytest.raises(ValueError):
        test_function()

    with pytest.raises(ValueError):
        test_function(id=1, path="test", other="other")


def test_feedback_score_in_range():
    assert SpanFeedback(score=0.5).score == 0.5
    assert SpanFeedback(score=0).score == 0
    assert SpanFeedback(score=1).score == 1
    with pytest.raises(ValidationError):
        SpanFeedback(score=1.1)
    with pytest.raises(ValidationError):
        SpanFeedback(score=-0.1)



from functools import wraps
from typing import Callable


def validate_id_xor_path(func: Callable) -> Callable:
    """
    A decorator to validate that either 'id' or 'path' is provided as an argument.
    Raises ValueError if neither is provided.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        if "id" not in kwargs and "path" not in kwargs:
            raise ValueError("Either 'id' or 'path' must be provided as an argument.")
        elif "id" in kwargs and "path" in kwargs:
            raise ValueError(
                "Only one of 'id' or 'path' should be provided as an argument."
            )
        return func(*args, **kwargs)

    return wrapper



import pytest
import time
from opperai import Client
from opperai.types.indexes import Document, Filter
from contextlib import contextmanager


@contextmanager
def index(name, _client: Client):
    idx = _client.indexes.create(name=name)
    try:
        yield idx
    finally:
        _client.indexes.delete(id=idx.id)


def test_create_index(vcr_cassette, client: Client):
    with index("test_create_index", client) as idx_id:
        assert idx_id is not None


def test_delete_index(client: Client, vcr_cassette):
    idx = client.indexes.create(name="test_delete_index")
    client.indexes.delete(id=idx.id)
    assert client.indexes.get(id=idx.id) is None


def test_get_by_id(vcr_cassette, client: Client):
    with index("test_get_by_id", client) as idx:
        idx_out = client.indexes.get(id=idx.id)
        assert idx_out.id == idx.id
        assert idx.name == "test_get_by_id"


def test_get_by_name(vcr_cassette, client: Client):
    with index("test_get_by_name", client) as idx:
        idx_out = client.indexes.get(name="test_get_by_name")
        assert idx_out.id == idx.id
        assert idx_out.name == "test_get_by_name"


def test_list_indexes(vcr_cassette, client: Client):
    idxs = client.indexes.list()
    assert len(idxs) == 0

    with index("test_list_indexes", client) as idx:
        idxs = client.indexes.list()
        assert len(idxs) == 1
        assert idxs[0].id == idx.id
        assert idxs[0].name == "test_list_indexes"

        with index("test_list_indexes_2", client) as idx_2:
            idxs = client.indexes.list()
            assert len(idxs) == 2
            assert idxs[1].id == idx_2.id
            assert idxs[1].name == "test_list_indexes_2"


def test_index_document(vcr_cassette, client: Client):
    with index("test_index_document", client) as idx:
        doc_in = Document(content="Hello", metadata={"source": "test"})
        doc_out = client.indexes.index(id=idx.id, doc=doc_in)
        assert doc_out.id is not None
        assert doc_out.uuid is not None
        assert doc_out.key is not None


def test_retrieve_document(vcr_cassette, client: Client):
    with index("test_retrieve_document", client) as idx:
        doc_in = Document(content="Hello", metadata={"source": "test"})
        client.indexes.index(id=idx.id, doc=doc_in)
        doc_retrieved = client.indexes.retrieve(id=idx.id, query="Hello", k=1)
        assert len(doc_retrieved) == 1
        assert doc_retrieved[0].content == "Hello"
        assert doc_retrieved[0].metadata == {"source": "test"}


@pytest.mark.skip(reason="Disabled due to pre signed urls")
def test_upload_file(client: Client, vcr_cassette):
    with index("test_upload_file", client) as idx:
        with open("/tmp/test.txt", "wb") as f:
            f.write(b"Hello")
            f.seek(0)
            client.indexes.upload_file(id=idx.id, file_path=f.name)
            time.sleep(0.5)  # wait for the file to be processed

        doc_retrieved = client.indexes.retrieve(id=idx.id, query="Hello", k=1)
        assert len(doc_retrieved) == 1
        assert doc_retrieved[0].content == "Hello"


def test_retrieve_filters(client: Client, vcr_cassette):
    with index("test_retrieve_filters", client) as idx:
        doc_in = Document(content="Bonjour", metadata={"source": "test"})
        client.indexes.index(id=idx.id, doc=doc_in)
        resp = client.indexes.retrieve(
            id=idx.id,
            query="test",
            k=1,
            filters=[Filter(key="source", operation="=", value="test")],
        )

        assert resp[0].content == "Bonjour"
        assert resp[0].metadata == {"source": "test"}



from typing import Generator, Optional

from opperai._http_clients import _http_client
from opperai.spans import get_current_span_id
from opperai.types import (
    ChatPayload,
    Function,
    FunctionResponse,
    StreamingChunk,
    validate_id_xor_path,
)
from http import HTTPStatus
from opperai.types.exceptions import (
    APIError,
    RateLimitError,
    StructuredGenerationError,
)


class Functions:
    def __init__(self, http_client: _http_client, default_model: str = None):
        self.http_client = http_client
        self.default_model = default_model

    def create(self, function: Function, update: bool = True, **kwargs) -> Function:
        fn = self.get(path=function.path)
        if fn is None:
            return self._create(function, **kwargs)
        elif update:
            function.id = fn.id
            return self.update(function, **kwargs)
        else:
            return fn

    def _create(self, function: Function, **kwargs) -> Function:
        if not function.model and self.default_model:
            function.model = self.default_model
        response = self.http_client.do_request(
            "POST",
            "/api/v1/functions",
            json={**function.model_dump(), **kwargs},
        )
        if response.status_code != HTTPStatus.OK:
            raise APIError(
                f"Failed to create function {function.path} with status {response.status_code}: {response.text}"
            )

        return Function.model_validate(response.json())

    def update(self, function: Function, **kwargs) -> Function:
        response = self.http_client.do_request(
            "POST",
            f"/api/v1/functions/{function.id}",
            json={**function.model_dump(), **kwargs},
        )
        if response.status_code != HTTPStatus.OK:
            raise APIError(
                f"Failed to update function `{function.path}` with status {response.status_code}: {response.text}"
            )

        return Function.model_validate(response.json())

    @validate_id_xor_path
    def get(self, id: str = None, path: str = None) -> Optional[Function]:
        if path is not None:
            if id is not None:
                raise ValueError("Only one of id or path should be provided")
            return self._get_by_path(path)
        elif id is not None:
            return self._get_by_id(id)
        else:
            return None

    def _get_by_path(self, function_path: str) -> Optional[Function]:
        response = self.http_client.do_request(
            "GET",
            f"/api/v1/functions/by_path/{function_path}",
        )
        if response.status_code == HTTPStatus.NOT_FOUND:
            return None
        if response.status_code != HTTPStatus.OK:
            raise APIError(
                f"Failed to get function {function_path} with status {response.status_code}"
            )

        return Function.model_validate(response.json())

    def _get_by_id(self, function_id: str) -> Optional[Function]:
        response = self.http_client.do_request(
            "GET",
            f"/api/v1/functions/{function_id}",
        )
        if response.status_code == HTTPStatus.NOT_FOUND:
            return None
        if response.status_code != HTTPStatus.OK:
            raise APIError(
                f"Failed to get function {function_id} with status {response.status_code}"
            )

        return Function.model_validate(response.json())

    @validate_id_xor_path
    def delete(self, id: str = None, path: str = None) -> bool:
        if path is not None:
            try:
                return self._delete_by_path(path)
            except APIError:
                pass
        elif id is not None:
            fn = self.get(id=id)
            if fn:
                return self._delete_by_path(fn.path)

    def _delete_by_path(self, function_path: str) -> bool:
        response = self.http_client.do_request(
            "DELETE",
            f"/api/v1/functions/by_path/{function_path}",
        )
        if response.status_code != HTTPStatus.OK:
            raise APIError(
                f"Failed to delete function {function_path} with status {response.status_code}"
            )
        return True

    def chat(
        self, function_path, data: ChatPayload, stream=False, **kwargs
    ) -> [FunctionResponse, Generator[StreamingChunk, None, None]]:
        if data.parent_span_uuid is None:
            data.parent_span_uuid = get_current_span_id()

        if stream:
            return self._chat_stream(function_path, data, **kwargs)
        serialized_data = data.model_dump()
        response = self.http_client.do_request(
            "POST",
            f"/v1/chat/{function_path}",
            json={**serialized_data, **kwargs},
        )

        if response.status_code == HTTPStatus.OK:
            return FunctionResponse.model_validate(response.json())
        elif response.status_code == HTTPStatus.TOO_MANY_REQUESTS:
            raise RateLimitError("Rate limit error: please retry in a few seconds")
        elif response.status_code == HTTPStatus.BAD_REQUEST:
            raise StructuredGenerationError(response.text)

        raise APIError(
            f"Failed to run function {function_path} with status {response.status_code}"
        )

    def _chat_stream(
        self, function_path, data: ChatPayload, **kwargs
    ) -> Generator[StreamingChunk, None, None]:
        serialized_data = data.model_dump()
        gen = self.http_client.stream(
            "POST",
            f"/v1/chat/{function_path}",
            json={**serialized_data, **kwargs},
            params={"stream": "True"},
        )
        for item in gen:
            yield StreamingChunk(**item)

    def flush_cache(self, id: int) -> bool:
        response = self.http_client.do_request(
            "DELETE",
            f"/api/v1/functions/{id}/cache",
        )
        if response.status_code != HTTPStatus.NO_CONTENT:
            raise APIError(
                f"Failed to flush cache for function with id={id} with status {response.status_code}"
            )

        return True



import asyncio
import contextlib
import inspect
import json
import os
import threading
from functools import wraps
from typing import List, get_args, get_origin, get_type_hints

from pydantic import BaseModel

from opperai import AsyncClient, Client
from opperai.spans import _current_span_id
from opperai.types import CacheConfiguration, ChatPayload, Function, Message

from ...utils import convert_function_call_to_json
from ._schemas import get_output_schema

_thread_local = threading.local()


@contextlib.contextmanager
def span_id_context():
    _thread_local.span_id = None
    try:
        yield
    finally:
        del _thread_local.span_id


def get_last_span_id() -> str:
    """Retrieve the last span ID from thread-local storage."""
    return getattr(_thread_local, "span_id", None)


def fn(
    _func=None,
    *,
    path=None,
    client=None,
    json_encoder=None,
    model=None,
    few_shot=None,
    few_shot_count=None,
    cache_config: CacheConfiguration = None,
):
    def decorator(func):
        func_path = path or func.__name__
        setup_done = False
        sync_client = None
        c = None

        def setup():
            nonlocal setup_done, sync_client, c
            if setup_done:
                return

            if isinstance(client, AsyncClient):
                sync_client = Client(api_key=client.api_key, api_url=client.api_url)
            elif isinstance(client, Client):
                sync_client = client
            else:
                sync_client = Client()

            use_few_shot = few_shot or os.environ.get("OPPER_USE_FEW_SHOT", False)
            function = Function(
                path=func_path,
                description=func.__doc__,
                instructions=f"Operation: {func.__name__}\n\nOperation description: {func.__doc__}",
                out_schema=get_output_schema(func),
                few_shot=use_few_shot,
            )
            if use_few_shot:
                function.use_semantic_search = True
                function.few_shot_count = few_shot_count or 2
            if cache_config:
                function.cache_configuration = cache_config

            function.model = (
                model if model else os.environ.get("OPPER_DEFAULT_MODEL", None)
            )

            sync_client.functions.create(function)

            if asyncio.iscoroutinefunction(func):
                c = AsyncClient() or client
            else:
                c = sync_client or client

            setup_done = True

        @wraps(func)
        async def async_wrapper(*args, **kwargs):
            setup()
            input = convert_function_call_to_json(func, *args, **kwargs)
            payload = ChatPayload(
                parent_span_uuid=_current_span_id.get(),
                messages=[
                    Message(role="user", content=json.dumps(input, cls=json_encoder))
                ],
            )

            response = await c.functions.chat(func_path, payload)
            answer = response.json_payload

            _thread_local.span_id = response.span_id
            return_type = get_type_hints(func).get("return")

            if return_type is not None:
                if inspect.isclass(return_type) and issubclass(return_type, BaseModel):
                    answer = return_type(**answer)
                elif (
                    (get_origin(return_type) == list or get_origin(return_type) == List)
                    and inspect.isclass(get_args(return_type)[0])
                    and issubclass(get_args(return_type)[0], BaseModel)
                ):
                    answer = [get_args(return_type)[0](**item) for item in answer]
            return answer

        def sync_wrapper(*args, **kwargs):
            setup()
            input = convert_function_call_to_json(func, *args, **kwargs)
            payload = ChatPayload(
                parent_span_uuid=_current_span_id.get(),
                messages=[
                    Message(role="user", content=json.dumps(input, cls=json_encoder))
                ],
            )
            response = c.functions.chat(func_path, payload)
            answer = response.json_payload
            _thread_local.span_id = response.span_id

            return_type = get_type_hints(func).get("return")
            if return_type is not None:
                if inspect.isclass(return_type) and issubclass(return_type, BaseModel):
                    answer = return_type(**answer)
                elif (
                    (get_origin(return_type) == list or get_origin(return_type) == List)
                    and inspect.isclass(get_args(return_type)[0])
                    and issubclass(get_args(return_type)[0], BaseModel)
                ):
                    answer = [get_args(return_type)[0](**item) for item in answer]
            return answer

        return async_wrapper if asyncio.iscoroutinefunction(func) else sync_wrapper

    if _func is None:
        return decorator
    else:
        return decorator(_func)



# ruff: noqa: F401
from typing import Any, Dict, List, Optional, Union
from uuid import UUID
from .spans import SpanFeedback

from pydantic import BaseModel, ConfigDict, Field

from .validators import validate_id_xor_path


class Message(BaseModel):
    role: str
    content: str


class ChatPayload(BaseModel):
    model_config: ConfigDict = ConfigDict(extra="allow")
    parent_span_uuid: Optional[str] = None

    messages: List[Message]


class FileMetadata(BaseModel):
    file_name: str


class ContextData(BaseModel):
    index_id: int
    content: str
    metadata: Union[Optional[dict], FileMetadata] = Field(default=None)


class StreamingChunk(BaseModel):
    span_id: Optional[str] = None
    delta: Optional[str] = None
    error: Optional[str] = None
    context: Optional[List[ContextData]] = None


class FunctionResponse(BaseModel):
    span_id: Optional[str] = None
    message: Optional[str] = None
    json_payload: Optional[Union[dict, List, Any]] = None
    error: Optional[str] = None
    context: Optional[List[ContextData]] = None
    cached: Optional[bool] = None


class CacheConfiguration(BaseModel):
    exact_match_cache_ttl: Optional[int] = None
    semantic_cache_threshold: Optional[float] = None
    semantic_cache_ttl: Optional[int] = None


class Function(BaseModel):
    id: Optional[int] = None
    path: str = Field(
        ...,
        pattern=r"^[a-zA-Z0-9_]+(\/[a-zA-Z0-9_-]+)*$",
        description="Path should not contain characters that could break URLs.",
    )
    description: str
    input_schema: Optional[Dict[str, Any]] = None
    out_schema: Optional[Dict[str, Any]] = None
    instructions: str
    model: Optional[str] = None
    index_ids: Optional[List[int]] = []
    use_semantic_search: Optional[bool] = None
    few_shot: Optional[bool] = None
    few_shot_count: Optional[int] = None
    cache_configuration: Optional[CacheConfiguration] = None



import json
from datetime import datetime
from uuid import UUID
from pydantic import BaseModel


from inspect import signature


def convert_function_call_to_json(func, *args, **kwargs):
    input_data = dict(zip(signature(func).parameters, args))
    input_data.update(kwargs)
    for key, value in input_data.items():
        if isinstance(value, BaseModel):
            input_data[key] = value.model_dump()
        elif isinstance(value, list) and all(isinstance(v, BaseModel) for v in value):
            input_data[key] = [v.model_dump() for v in value]
    return input_data


class DateTimeEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.isoformat()
        if isinstance(obj, UUID):
            return str(obj)
        # Let the base class default method raise the TypeError
        return super().default(obj)



from datetime import datetime
from opperai.types.spans import Span
from opperai import Client


def test_crud(client: Client, vcr_cassette):
    root_uuid = "bcf1b3b4-0b3b-4b3b-8b3b-0b3b4b3b4b3b"
    root_start_time = datetime(2021, 1, 1, 0, 0, 0, 0)
    root_end_time = datetime(2022, 1, 1, 0, 0, 0, 0)

    child_uuid = "cfc1c3c4-0c3c-4c3c-8c3c-0c3c4c3c4c3c"
    child_start_time = datetime(2021, 1, 1, 0, 0, 0, 0)
    child_end_time = datetime(2021, 12, 1, 0, 0, 0, 0)

    root_span = Span(
        uuid=root_uuid,
        project="project",
        name="name",
        input="input",
        start_time=root_start_time,
    )

    client.spans.create(root_span)

    child_span = Span(
        uuid=child_uuid,
        parent_uuid=root_span.uuid,
        project="project",
        name="name",
        input="input",
        start_time=child_start_time,
    )

    client.spans.create(child_span)

    client.spans.update(
        child_span.uuid,
        end_time=child_end_time,
        output="output",
    )

    client.spans.update(
        root_span.uuid,
        end_time=root_end_time,
        output="output",
    )

    assert client.spans.delete(child_span.uuid)
    assert client.spans.delete(root_span.uuid)



from datetime import datetime
import pytest
from opperai.types.spans import Span, SpanFeedback
from opperai import AsyncClient
from contextlib import asynccontextmanager


@asynccontextmanager
async def span(desc: Span, _client: AsyncClient):
    _span = await _client.spans.create(desc)
    try:
        yield _span
    finally:
        await _client.spans.delete(_span.uuid)


@pytest.mark.asyncio(scope="module")
async def test_async_crud(aclient: AsyncClient, vcr_cassette):
    root_uuid = "bcf1b3b4-0b3b-4b3b-8b3b-0b3b4b3b4112"
    root_start_time = datetime(2021, 1, 1, 0, 0, 0, 0)
    root_end_time = datetime(2022, 1, 1, 0, 0, 0, 0)

    child_uuid = "cfc1c3c4-0c3c-4c3c-8c3c-0c3c4c31cc12"
    child_start_time = datetime(2021, 1, 1, 0, 0, 0, 0)
    child_end_time = datetime(2021, 12, 1, 0, 0, 0, 0)

    root_span = Span(
        uuid=root_uuid,
        project="project",
        name="name",
        input="input",
        start_time=root_start_time,
    )

    await aclient.spans.create(root_span)

    child_span = Span(
        uuid=child_uuid,
        parent_uuid=root_span.uuid,
        project="project",
        name="name",
        input="input",
        start_time=child_start_time,
    )

    await aclient.spans.create(child_span)

    await aclient.spans.update(
        child_span.uuid,
        end_time=child_end_time,
        output="output",
    )

    await aclient.spans.update(
        root_span.uuid,
        end_time=root_end_time,
        output="output",
    )

    assert await aclient.spans.delete(child_span.uuid)
    assert await aclient.spans.delete(root_span.uuid)


@pytest.mark.asyncio(scope="module")
async def test_save_feedback(aclient: AsyncClient, vcr_cassette):
    async with span(
        Span(
            uuid="bcf1b3b4-0b3b-4b3b-8b3b-0b3b4b3b4133",
            project="project",
            name="name",
            input="input",
            start_time=datetime(2021, 1, 1, 0, 0, 0, 0),
        ),
        aclient,
    ) as _span:
        await aclient.spans.save_feedback(
            _span.uuid, SpanFeedback(dimension="dim", score=0.5, comment="comment")
        )



import pytest
from opperai import AsyncClient
from opperai.types import CacheConfiguration, ChatPayload, Function, Message
from opperai.types.exceptions import StructuredGenerationError
from contextlib import asynccontextmanager


@asynccontextmanager
async def _function(desc: Function, c: AsyncClient):
    function = await c.functions.create(desc)
    yield function
    await c.functions.delete(id=function.id)


@pytest.mark.asyncio(scope="module")
async def test_create_function(vcr_cassette, aclient: AsyncClient):
    async with _function(
        Function(
            path="test/sdk/test_create_function_async",
            description="Test function",
            instructions="Do something",
        ),
        aclient,
    ) as function:
        assert function is not None
        assert function.path == "test/sdk/test_create_function_async"


@pytest.mark.asyncio(scope="module")
async def test_get_by_id(aclient: AsyncClient, vcr_cassette):
    async with _function(
        Function(
            path="test/sdk/test_get_by_id",
            description="Test function",
            instructions="Do something",
        ),
        aclient,
    ) as function:
        f_by_id = await aclient.functions.get(id=function.id)
        assert f_by_id.path == "test/sdk/test_get_by_id"


@pytest.mark.asyncio(scope="module")
async def test_get_by_path(aclient: AsyncClient, vcr_cassette):
    async with _function(
        Function(
            path="test/sdk/test_get_by_path",
            description="Test function",
            instructions="Do something",
        ),
        aclient,
    ) as _:
        f_by_path = await aclient.functions.get(path="test/sdk/test_get_by_path")
        assert f_by_path.path == "test/sdk/test_get_by_path"


@pytest.mark.asyncio(scope="module")
async def test_get(aclient: AsyncClient, vcr_cassette):
    async with _function(
        Function(
            path="test/sdk/test_get",
            description="Test function",
            instructions="Do something",
        ),
        aclient,
    ) as function:
        f_by_id = await aclient.functions.get(id=function.id)
        assert f_by_id.path == "test/sdk/test_get"

        f_by_path = await aclient.functions.get(path="test/sdk/test_get")
        assert f_by_path.path == "test/sdk/test_get"


@pytest.mark.asyncio(scope="module")
async def test_update_function(aclient: AsyncClient, vcr_cassette):
    async with _function(
        Function(
            path="test/sdk/test_update_function",
            description="Test function",
            instructions="Do something",
        ),
        aclient,
    ) as function:
        function.instructions = "Do something else"
        updated_function = await aclient.functions.update(function)
        assert updated_function.instructions == "Do something else"


@pytest.mark.asyncio(scope="module")
async def test_delete_function(aclient: AsyncClient, vcr_cassette):
    function = await aclient.functions.create(
        Function(
            path="test/sdk/test_delete_function",
            description="Test function",
            instructions="Do something",
        )
    )
    await aclient.functions.delete(id=function.id)
    f = await aclient.functions.get(id=function.id)
    assert f is None


@pytest.mark.asyncio(scope="module")
async def test_async_chat(aclient: AsyncClient, vcr_cassette):
    async with _function(
        Function(
            path="test/sdk/test_async_chat",
            description="Translate to French",
            instructions="Translate to French",
        ),
        aclient,
    ) as function:
        resp = await aclient.functions.chat(
            function.path, ChatPayload(messages=[Message(role="user", content="hello")])
        )

        assert "bonjour" in resp.message.lower()


@pytest.mark.asyncio(scope="module")
async def test_async_chat_stream(aclient: AsyncClient, vcr_cassette):
    async with _function(
        Function(
            path="test/sdk/test_async_chat_stream",
            description="Translate to French",
            instructions="Translate to French",
        ),
        aclient,
    ) as function:
        gen = await aclient.functions.chat(
            function.path,
            ChatPayload(messages=[Message(role="user", content="hello")]),
            stream=True,
        )

        resp = "".join([message.delta async for message in gen if message.delta])

        assert "bonjour" in resp.lower()


@pytest.mark.asyncio(scope="module")
async def test_create_function_with_cache(aclient: AsyncClient, vcr_cassette):
    async with _function(
        Function(
            path="test/sdk/test_create_function_with_cache_async",
            description="Test function",
            instructions="Do something",
            cache_configuration=CacheConfiguration(exact_match_cache_ttl=10),
        ),
        aclient,
    ) as function:
        res = await aclient.functions.chat(
            function.path, ChatPayload(messages=[Message(role="user", content="hello")])
        )
        print(res)
        assert not res.cached

        res = await aclient.functions.chat(
            function.path, ChatPayload(messages=[Message(role="user", content="hello")])
        )
        print(res)
        assert res.cached


@pytest.mark.asyncio(scope="module")
async def test_create_function_with_cache_flush(aclient: AsyncClient, vcr_cassette):
    async with _function(
        Function(
            path="test/sdk/test_create_function_with_cache_async_flush",
            description="Test function",
            instructions="Do something",
            cache_configuration=CacheConfiguration(exact_match_cache_ttl=10),
        ),
        aclient,
    ) as function:
        res = await aclient.functions.chat(
            function.path, ChatPayload(messages=[Message(role="user", content="hello")])
        )
        print(res)
        assert not res.cached

        res = await aclient.functions.chat(
            function.path, ChatPayload(messages=[Message(role="user", content="hello")])
        )
        print(res)
        assert res.cached

        await aclient.functions.flush_cache(id=function.id)
        res = await aclient.functions.chat(
            function.path, ChatPayload(messages=[Message(role="user", content="hello")])
        )
        print(res)
        assert not res.cached


@pytest.mark.asyncio(scope="module")
async def test_failed_structured_generation(aclient: AsyncClient, vcr_cassette):
    async with _function(
        Function(
            model="mistral/mistral-tiny-eu",
            path="test/sdk/test_failed_structured_generation",
            description="test structured generation exception",
            instructions="You translate the incoming text to french",
            out_schema={
                "type": "object",
                "properties": {
                    "universityName": {"type": "string"},
                    "location": {"type": "string"},
                    "departments": {
                        "type": "array",
                        "items": {
                            "type": "object",
                            "properties": {
                                "departmentName": {"type": "string"},
                                "head": {"type": "string"},
                                "courses": {
                                    "type": "array",
                                    "items": {
                                        "type": "object",
                                        "properties": {
                                            "courseId": {"type": "string"},
                                            "courseName": {"type": "string"},
                                            "credits": {
                                                "type": "integer",
                                                "minimum": 1,
                                                "maximum": 10,
                                            },
                                        },
                                        "required": [
                                            "courseId",
                                            "courseName",
                                            "credits",
                                        ],
                                    },
                                },
                            },
                            "required": ["departmentName", "head", "courses"],
                        },
                    },
                },
                "required": ["universityName", "location", "departments"],
            },
        ),
        aclient,
    ) as function:
        with pytest.raises(StructuredGenerationError):
            await aclient.functions.chat(
                function.path,
                ChatPayload(messages=[Message(role="user", content="hello")]),
            )



import pytest
import asyncio
from opperai import AsyncClient
from opperai.types.indexes import Document, Filter
from contextlib import asynccontextmanager


@asynccontextmanager
async def _index(name, _client: AsyncClient):
    index = await _client.indexes.create(name=name)
    try:
        yield index
    finally:
        await _client.indexes.delete(id=index.id)


@pytest.mark.asyncio(scope="module")
async def test_create_index(aclient: AsyncClient, vcr_cassette):
    async with _index("test_create_index", aclient) as index:
        assert index is not None


@pytest.mark.asyncio(scope="module")
async def test_delete_index(aclient: AsyncClient, vcr_cassette):
    async with _index("test_delete_index", aclient) as index:
        await aclient.indexes.delete(id=index.id)
        assert await aclient.indexes.get(id=index.id) is None


@pytest.mark.asyncio(scope="module")
async def test_get_by_id(vcr_cassette, aclient: AsyncClient):
    async with _index("test_get_by_id", aclient) as index:
        idx_out = await aclient.indexes.get(id=index.id)
        assert idx_out.id == index.id
        assert idx_out.name == "test_get_by_id"


@pytest.mark.asyncio(scope="module")
async def test_get_by_name(vcr_cassette, aclient: AsyncClient):
    async with _index("test_get_by_name", aclient) as index:
        idx_out = await aclient.indexes.get(name="test_get_by_name")
        assert idx_out.id == index.id
        assert idx_out.name == "test_get_by_name"


@pytest.mark.asyncio(scope="module")
async def test_list_indexes(vcr_cassette, aclient: AsyncClient):
    idxs = await aclient.indexes.list()
    assert len(idxs) == 0

    async with _index("test_list_indexes", aclient) as index:
        idxs = await aclient.indexes.list()
        assert len(idxs) == 1
        assert idxs[0].id == index.id
        assert idxs[0].name == "test_list_indexes"

        async with _index("test_list_indexes_2", aclient) as index_2:
            idxs = await aclient.indexes.list()
            assert len(idxs) == 2
            assert idxs[1].id == index_2.id
            assert idxs[1].name == "test_list_indexes_2"


@pytest.mark.asyncio(scope="module")
async def test_index_document(vcr_cassette, aclient: AsyncClient):
    async with _index("test_index_document", aclient) as index:
        doc_in = Document(content="Hello", metadata={"source": "test"})
        doc_out = await aclient.indexes.index(id=index.id, doc=doc_in)
        assert doc_out.id is not None
        assert doc_out.uuid is not None
        assert doc_out.key is not None


@pytest.mark.asyncio(scope="module")
async def test_retrieve_document(vcr_cassette, aclient: AsyncClient):
    async with _index("test_retrieve_document", aclient) as index:
        doc_in = Document(content="Hello", metadata={"source": "test"})
        await aclient.indexes.index(id=index.id, doc=doc_in)
        doc_retrieved = await aclient.indexes.retrieve(id=index.id, query="Hello", k=1)
        assert len(doc_retrieved) == 1
        assert doc_retrieved[0].content == "Hello"
        assert doc_retrieved[0].metadata == {"source": "test"}


@pytest.mark.skip(reason="Disabled due to pre signed urls")
@pytest.mark.asyncio(scope="module")
async def test_upload_file(aclient: AsyncClient, vcr_cassette):
    async with _index("test_upload_file", aclient) as index:
        with open("/tmp/test.txt", "wb") as f:
            f.write(b"Hello")
            f.seek(0)
            await aclient.indexes.upload_file(id=index.id, file_path=f.name)
            await asyncio.sleep(0.5)  # wait for the file to be processed

        doc_retrieved = await aclient.indexes.retrieve(id=index.id, query="Hello", k=1)
        assert len(doc_retrieved) == 1
        assert doc_retrieved[0].content == "Hello"


@pytest.mark.asyncio(scope="module")
async def test_retrieve_filters(aclient: AsyncClient, vcr_cassette):
    async with _index("test_retrieve_filters", aclient) as index:
        doc_in = Document(content="Bonjour", metadata={"source": "test"})
        await aclient.indexes.index(id=index.id, doc=doc_in)
        resp = await aclient.indexes.retrieve(
            id=index.id,
            query="test",
            k=1,
            filters=[Filter(key="source", operation="=", value="test")],
        )

        assert resp[0].content == "Bonjour"
        assert resp[0].metadata == {"source": "test"}
